@@ -5,6 +5,8 @@
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -30,7 +32,6 @@
 
 #include "config.h"
 
-#include <errno.h>
 #include <dlfcn.h>
 #include <fnmatch.h>
 #include <stdio.h>
@@ -44,13 +45,18 @@
 #include "cache_director.h"
 #include "cache_vcl.h"
 #include "vcli_serve.h"
+#include "vte.h"
+#include "vtim.h"
+#include "vcc_interface.h"
 
-const char * const VCL_TEMP_INIT = "init";
-const char * const VCL_TEMP_COLD = "cold";
-const char * const VCL_TEMP_WARM = "warm";
-const char * const VCL_TEMP_BUSY = "busy";
-const char * const VCL_TEMP_COOLING = "cooling";
-const char * const VCL_TEMP_LABEL = "label";
+const struct vcltemp VCL_TEMP_INIT[1] = {{ .name = "init", .is_cold = 1 }};
+const struct vcltemp VCL_TEMP_COLD[1] = {{ .name = "cold", .is_cold = 1 }};
+const struct vcltemp VCL_TEMP_WARM[1] = {{ .name = "warm", .is_warm = 1 }};
+const struct vcltemp VCL_TEMP_BUSY[1] = {{ .name = "busy", .is_warm = 1 }};
+const struct vcltemp VCL_TEMP_COOLING[1] = {{ .name = "cooling" }};
+
+// not really a temperature
+static const struct vcltemp VCL_TEMP_LABEL[1] = {{ .name = "label" }};
 
 /*
  * XXX: Presently all modifications to this list happen from the
@@ -63,9 +69,10 @@ struct lock		vcl_mtx;
 struct vcl		*vcl_active; /* protected by vcl_mtx */
 
 static struct vrt_ctx ctx_cli;
-static unsigned handling_cli;
+static struct wrk_vpi wrk_vpi_cli;
 static struct ws ws_cli;
 static uintptr_t ws_snapshot_cli;
+static struct vsl_log vsl_cli;
 
 /*--------------------------------------------------------------------*/
 
@@ -75,7 +82,9 @@ VCL_Bo2Ctx(struct vrt_ctx *ctx, struct busyobj *bo)
 
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
 	CHECK_OBJ_NOTNULL(bo, BUSYOBJ_MAGIC);
+	CHECK_OBJ_NOTNULL(bo->wrk, WORKER_MAGIC);
 	ctx->vcl = bo->vcl;
+	ctx->syntax = ctx->vcl->conf->syntax;
 	ctx->vsl = bo->vsl;
 	ctx->http_bereq = bo->bereq;
 	ctx->http_beresp = bo->beresp;
@@ -83,6 +92,9 @@ VCL_Bo2Ctx(struct vrt_ctx *ctx, struct busyobj *bo)
 	ctx->sp = bo->sp;
 	ctx->now = bo->t_prev;
 	ctx->ws = bo->ws;
+	ctx->vpi = bo->wrk->vpi;
+	ctx->vpi->handling = 0;
+	ctx->vpi->trace = bo->trace;
 }
 
 void
@@ -91,79 +103,160 @@ VCL_Req2Ctx(struct vrt_ctx *ctx, struct req *req)
 
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
 	CHECK_OBJ_NOTNULL(req, REQ_MAGIC);
+	CHECK_OBJ_NOTNULL(req->doclose, STREAM_CLOSE_MAGIC);
 
 	ctx->vcl = req->vcl;
+	ctx->syntax = ctx->vcl->conf->syntax;
 	ctx->vsl = req->vsl;
 	ctx->http_req = req->http;
-	ctx->http_req_top = req->top->http;
+	CHECK_OBJ_NOTNULL(req->top, REQTOP_MAGIC);
+	ctx->http_req_top = req->top->topreq->http;
 	ctx->http_resp = req->resp;
 	ctx->req = req;
 	ctx->sp = req->sp;
 	ctx->now = req->t_prev;
 	ctx->ws = req->ws;
+	ctx->vpi = req->wrk->vpi;
+	ctx->vpi->handling = 0;
+	ctx->vpi->trace = req->trace;
 }
 
 /*--------------------------------------------------------------------*/
 
-static struct vrt_ctx *
-vcl_get_ctx(unsigned method, int msg)
+struct vrt_ctx *
+VCL_Get_CliCtx(int msg)
 {
 
 	ASSERT_CLI();
-	AZ(ctx_cli.handling);
 	INIT_OBJ(&ctx_cli, VRT_CTX_MAGIC);
-	handling_cli = 0;
-	ctx_cli.handling = &handling_cli;
-	ctx_cli.method = method;
+	INIT_OBJ(&wrk_vpi_cli, WRK_VPI_MAGIC);
+	ctx_cli.vpi = &wrk_vpi_cli;
+	wrk_vpi_cli.trace = FEATURE(FEATURE_TRACE);
+	ctx_cli.now = VTIM_real();
 	if (msg) {
 		ctx_cli.msg = VSB_new_auto();
 		AN(ctx_cli.msg);
+	} else {
+		ctx_cli.vsl = &vsl_cli;
 	}
 	ctx_cli.ws = &ws_cli;
 	WS_Assert(ctx_cli.ws);
 	return (&ctx_cli);
 }
 
-static void
-vcl_rel_ctx(struct vrt_ctx **ctx)
+/*
+ * releases CLI ctx
+ *
+ * returns finished error msg vsb if VCL_Get_CliCtx(1) was called
+ *
+ * caller needs to VSB_destroy a non-NULL return value
+ *
+ */
+struct vsb *
+VCL_Rel_CliCtx(struct vrt_ctx **ctx)
 {
+	struct vsb *r = NULL;
 
 	ASSERT_CLI();
 	assert(*ctx == &ctx_cli);
-	AN((*ctx)->handling);
-	if (ctx_cli.msg)
-		VSB_destroy(&ctx_cli.msg);
+	AN((*ctx)->vpi);
+	if (ctx_cli.msg) {
+		TAKE_OBJ_NOTNULL(r, &ctx_cli.msg, VSB_MAGIC);
+		AZ(VSB_finish(r));
+	}
+	if (ctx_cli.vsl)
+		VSL_Flush(ctx_cli.vsl, 0);
 	WS_Assert(ctx_cli.ws);
-	WS_Reset(&ws_cli, ws_snapshot_cli);
+	WS_Rollback(&ws_cli, ws_snapshot_cli);
 	INIT_OBJ(*ctx, VRT_CTX_MAGIC);
 	*ctx = NULL;
+
+	return (r);
 }
 
 /*--------------------------------------------------------------------*/
 
+/* VRT_fail() can be called
+ * - from the vcl sub via a vmod
+ * - via a PRIV_TASK .fini callback
+ *
+ * if this happens during init, we fail it
+ * if during fini, we ignore, because otherwise VMOD authors would be forced to
+ * handle VCL_MET_FINI specifically everywhere.
+ */
+
 static int
-vcl_send_event(VRT_CTX, enum vcl_event_e ev)
+vcl_event_handling(VRT_CTX)
 {
-	int r;
+	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
+
+	if (ctx->vpi->handling == 0)
+		return (0);
+
+	assert(ctx->vpi->handling == VCL_RET_FAIL);
+
+	if (ctx->method == VCL_MET_INIT)
+		return (1);
+
+	/*
+	 * EVENT_WARM / EVENT_COLD: method == 0
+	 * must not set handling
+	 */
+	assert(ctx->method == VCL_MET_FINI);
+
+	ctx->vpi->handling = 0;
+	VRT_fail(ctx, "VRT_fail() from vcl_fini{} has no effect");
+	return (0);
+}
+
+static int
+vcl_send_event(struct vcl *vcl, enum vcl_event_e ev, struct vsb **msg)
+{
+	int r, havemsg;
+	unsigned method = 0;
+	struct vrt_ctx *ctx;
 
 	ASSERT_CLI();
-	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
-	CHECK_OBJ_NOTNULL(ctx->vcl, VCL_MAGIC);
-	CHECK_OBJ_NOTNULL(ctx->vcl->conf, VCL_CONF_MAGIC);
-	assert(ev == VCL_EVENT_LOAD ||
-	       ev == VCL_EVENT_WARM ||
-	       ev == VCL_EVENT_COLD ||
-	       ev == VCL_EVENT_DISCARD);
-	AN(ctx->handling);
-	*ctx->handling = 0;
+	ASSERT_VCL_ACTIVE();
+
+	CHECK_OBJ_NOTNULL(vcl, VCL_MAGIC);
+	CHECK_OBJ_NOTNULL(vcl->conf, VCL_CONF_MAGIC);
+	AN(msg);
+	AZ(*msg);
+
+	switch (ev) {
+	case VCL_EVENT_LOAD:
+		method = VCL_MET_INIT;
+		/* FALLTHROUGH */
+	case VCL_EVENT_WARM:
+		havemsg = 1;
+		break;
+	case VCL_EVENT_DISCARD:
+		method = VCL_MET_FINI;
+		/* FALLTHROUGH */
+	case VCL_EVENT_COLD:
+		havemsg = 0;
+		break;
+	default:
+		WRONG("vcl_event");
+	}
+
+	ctx = VCL_Get_CliCtx(havemsg);
+
+	AN(ctx->vpi);
+	AZ(ctx->vpi->handling);
 	AN(ctx->ws);
 
-	if (ev == VCL_EVENT_LOAD || ev == VCL_EVENT_WARM)
-		AN(ctx->msg);
+	ctx->vcl = vcl;
+	ctx->syntax = ctx->vcl->conf->syntax;
+	ctx->method = method;
 
-	VCL_TaskEnter(ctx->vcl, cli_task_privs);
+	VCL_TaskEnter(cli_task_privs);
 	r = ctx->vcl->conf->event_vcl(ctx, ev);
-	VCL_TaskLeave(ctx->vcl, cli_task_privs);
+	VCL_TaskLeave(ctx, cli_task_privs);
+	r |= vcl_event_handling(ctx);
+
+	*msg = VCL_Rel_CliCtx(&ctx);
 
 	if (r && (ev == VCL_EVENT_COLD || ev == VCL_EVENT_DISCARD))
 		WRONG("A VMOD cannot fail COLD or DISCARD events");
@@ -190,67 +283,89 @@ vcl_find(const char *name)
 
 /*--------------------------------------------------------------------*/
 
+static void
+vcl_panic_conf(struct vsb *vsb, const struct VCL_conf *conf)
+{
+	unsigned u;
+	const struct vpi_ii *ii;
+
+	if (PAN_dump_struct(vsb, conf, VCL_CONF_MAGIC, "conf"))
+		return;
+	VSB_printf(vsb, "syntax = \"%u\",\n", conf->syntax);
+	VSB_cat(vsb, "srcname = {\n");
+	VSB_indent(vsb, 2);
+	for (u = 0; u < conf->nsrc; ++u)
+		VSB_printf(vsb, "[%u] = \"%s\",\n", u, conf->srcname[u]);
+	VSB_indent(vsb, -2);
+	VSB_cat(vsb, "},\n");
+	VSB_cat(vsb, "instances = {\n");
+	VSB_indent(vsb, 2);
+	ii = conf->instance_info;
+	while (ii != NULL && ii->p != NULL) {
+		VSB_printf(vsb, "\"%s\" = %p,\n", ii->name,
+		    (const void *)*(const uintptr_t *)ii->p);
+		ii++;
+	}
+	VSB_indent(vsb, -2);
+	VSB_cat(vsb, "},\n");
+	VSB_indent(vsb, -2);
+	VSB_cat(vsb, "},\n");
+}
+
 void
-VCL_Panic(struct vsb *vsb, const struct vcl *vcl)
+VCL_Panic(struct vsb *vsb, const char *nm, const struct vcl *vcl)
 {
-	int i;
 
 	AN(vsb);
-	if (vcl == NULL)
+	if (PAN_dump_struct(vsb, vcl, VCL_MAGIC, "vcl[%s]", nm))
 		return;
-	VSB_printf(vsb, "vcl = {\n");
-	VSB_indent(vsb, 2);
-	PAN_CheckMagic(vsb, vcl, VCL_MAGIC);
 	VSB_printf(vsb, "name = \"%s\",\n", vcl->loaded_name);
 	VSB_printf(vsb, "busy = %u,\n", vcl->busy);
 	VSB_printf(vsb, "discard = %u,\n", vcl->discard);
 	VSB_printf(vsb, "state = %s,\n", vcl->state);
-	VSB_printf(vsb, "temp = %s,\n", (const volatile char *)vcl->temp);
-	VSB_printf(vsb, "conf = {\n");
-	VSB_indent(vsb, 2);
-	if (vcl->conf == NULL) {
-		VSB_printf(vsb, "conf = NULL\n");
-	} else {
-		PAN_CheckMagic(vsb, vcl->conf, VCL_CONF_MAGIC);
-		VSB_printf(vsb, "srcname = {\n");
-		VSB_indent(vsb, 2);
-		for (i = 0; i < vcl->conf->nsrc; ++i)
-			VSB_printf(vsb, "\"%s\",\n", vcl->conf->srcname[i]);
-		VSB_indent(vsb, -2);
-		VSB_printf(vsb, "},\n");
-	}
+	VSB_printf(vsb, "temp = %s,\n", vcl->temp ? vcl->temp->name : "(null)");
+	vcl_panic_conf(vsb, vcl->conf);
 	VSB_indent(vsb, -2);
-	VSB_printf(vsb, "},\n");
-	VSB_indent(vsb, -2);
-	VSB_printf(vsb, "},\n");
+	VSB_cat(vsb, "},\n");
 }
 
 /*--------------------------------------------------------------------*/
 
 void
-vcl_get(struct vcl **vcc, struct vcl *vcl)
+VCL_Update(struct vcl **vcc, struct vcl *vcl)
 {
+	struct vcl *old;
+
 	AN(vcc);
 
+	old = *vcc;
+	*vcc = NULL;
+
+	CHECK_OBJ_ORNULL(old, VCL_MAGIC);
+	ASSERT_VCL_ACTIVE();
+
 	Lck_Lock(&vcl_mtx);
+	if (old != NULL) {
+		assert(old->busy > 0);
+		old->busy--;
+	}
+
 	if (vcl == NULL)
 		vcl = vcl_active; /* Sample vcl_active under lock to avoid
 				   * race */
 	CHECK_OBJ_NOTNULL(vcl, VCL_MAGIC);
 	if (vcl->label == NULL) {
-		AN(strcmp(vcl->state, VCL_TEMP_LABEL));
+		AN(strcmp(vcl->state, VCL_TEMP_LABEL->name));
 		*vcc = vcl;
 	} else {
-		AZ(strcmp(vcl->state, VCL_TEMP_LABEL));
+		AZ(strcmp(vcl->state, VCL_TEMP_LABEL->name));
 		*vcc = vcl->label;
 	}
 	CHECK_OBJ_NOTNULL(*vcc, VCL_MAGIC);
 	AZ((*vcc)->discard);
 	(*vcc)->busy++;
 	Lck_Unlock(&vcl_mtx);
-	AZ(errno=pthread_rwlock_rdlock(&(*vcc)->temp_rwl));
-	assert(VCL_WARM(*vcc));
-	AZ(errno=pthread_rwlock_unlock(&(*vcc)->temp_rwl));
+	assert((*vcc)->temp->is_warm);
 }
 
 /*--------------------------------------------------------------------*/
@@ -262,6 +377,7 @@ vcl_iterdir(struct cli *cli, const char *pat, const struct vcl *vcl,
 	int i, found = 0;
 	struct vcldir *vdir;
 
+	Lck_AssertHeld(&vcl_mtx);
 	VTAILQ_FOREACH(vdir, &vcl->director_list, list) {
 		if (fnmatch(pat, vdir->cli_name, 0))
 			continue;
@@ -283,6 +399,7 @@ VCL_IterDirector(struct cli *cli, const char *pat,
 	struct vcl *vcl;
 
 	ASSERT_CLI();
+	ASSERT_VCL_ACTIVE();
 	vsb = VSB_new_auto();
 	AN(vsb);
 	if (pat == NULL || *pat == '\0' || !strcmp(pat, "*")) {
@@ -327,29 +444,26 @@ vcl_BackendEvent(const struct vcl *vcl, enum vcl_event_e e)
 	CHECK_OBJ_NOTNULL(vcl, VCL_MAGIC);
 	AZ(vcl->busy);
 
+	Lck_Lock(&vcl_mtx);
 	VTAILQ_FOREACH(vdir, &vcl->director_list, list)
 		VDI_Event(vdir->dir, e);
+	Lck_Unlock(&vcl_mtx);
 }
 
 static void
-vcl_KillBackends(struct vcl *vcl)
+vcl_KillBackends(const struct vcl *vcl)
 {
-	struct vcldir *vdir;
+	struct vcldir *vdir, *vdir2;
 
 	CHECK_OBJ_NOTNULL(vcl, VCL_MAGIC);
-	AZ(vcl->busy);
-	assert(VTAILQ_EMPTY(&vcl->ref_list));
-	while (1) {
-		vdir = VTAILQ_FIRST(&vcl->director_list);
-		if (vdir == NULL)
-			break;
-		VTAILQ_REMOVE(&vcl->director_list, vdir, list);
-		REPLACE(vdir->cli_name, NULL);
-		AN(vdir->methods->destroy);
-		vdir->methods->destroy(vdir->dir);
-		FREE_OBJ(vdir->dir);
-		FREE_OBJ(vdir);
-	}
+	assert(vcl->temp == VCL_TEMP_COLD || vcl->temp == VCL_TEMP_INIT);
+	/*
+	 * Unlocked because no further directors can be added, and the
+	 * remaining ones need to be able to remove themselves.
+	 */
+	VTAILQ_FOREACH_SAFE(vdir, &vcl->director_list, list, vdir2)
+		VDI_Event(vdir->dir, VCL_EVENT_DISCARD);
+	assert(VTAILQ_EMPTY(&vcl->director_list));
 }
 
 /*--------------------------------------------------------------------*/
@@ -360,6 +474,8 @@ VCL_Open(const char *fn, struct vsb *msg)
 	struct vcl *vcl;
 	void *dlh;
 	struct VCL_conf const *cnf;
+	const char *dlerr;
+	int err;
 
 	AN(fn);
 	AN(msg);
@@ -371,22 +487,32 @@ VCL_Open(const char *fn, struct vsb *msg)
 #endif
 	dlh = dlopen(fn, RTLD_NOW | RTLD_LOCAL);
 	if (dlh == NULL) {
-		VSB_printf(msg, "Could not load compiled VCL.\n");
-		VSB_printf(msg, "\tdlopen() = %s\n", dlerror());
+		err = errno;
+		dlerr = dlerror();
+		VSB_cat(msg, "Could not load compiled VCL.\n");
+		if (dlerr != NULL)
+			VSB_printf(msg, "\tdlopen() = %s\n", dlerr);
+		if (err) {
+			VSB_printf(msg, "\terror = %s (%d)\n",
+			    strerror(err), err);
+		}
+		VSB_cat(msg, "\thint: check for \"noexec\" mount\n");
+		VSB_cat(msg, "\thint: check \"vmod_path\" parameter\n");
 		return (NULL);
 	}
 	cnf = dlsym(dlh, "VCL_conf");
 	if (cnf == NULL) {
-		VSB_printf(msg, "Compiled VCL lacks metadata.\n");
+		VSB_cat(msg, "Compiled VCL lacks metadata.\n");
 		(void)dlclose(dlh);
 		return (NULL);
 	}
 	if (cnf->magic != VCL_CONF_MAGIC) {
-		VSB_printf(msg, "Compiled VCL has mangled metadata.\n");
+		VSB_cat(msg, "Compiled VCL has mangled metadata.\n");
 		(void)dlclose(dlh);
 		return (NULL);
 	}
-	if (cnf->syntax < heritage.min_vcl || cnf->syntax > heritage.max_vcl) {
+	if (cnf->syntax < heritage.min_vcl_version ||
+	    cnf->syntax > heritage.max_vcl_version) {
 		VSB_printf(msg, "Compiled VCL version (%.1f) not supported.\n",
 		    .1 * cnf->syntax);
 		(void)dlclose(dlh);
@@ -394,7 +520,6 @@ VCL_Open(const char *fn, struct vsb *msg)
 	}
 	ALLOC_OBJ(vcl, VCL_MAGIC);
 	AN(vcl);
-	AZ(errno=pthread_rwlock_init(&vcl->temp_rwl, NULL));
 	vcl->dlh = dlh;
 	vcl->conf = cnf;
 	return (vcl);
@@ -405,16 +530,14 @@ VCL_Close(struct vcl **vclp)
 {
 	struct vcl *vcl;
 
-	CHECK_OBJ_NOTNULL(*vclp, VCL_MAGIC);
-	vcl = *vclp;
-	*vclp = NULL;
+	TAKE_OBJ_NOTNULL(vcl, vclp, VCL_MAGIC);
+	assert(VTAILQ_EMPTY(&vcl->filters));
 	AZ(dlclose(vcl->dlh));
-	AZ(errno=pthread_rwlock_destroy(&vcl->temp_rwl));
 	FREE_OBJ(vcl);
 }
 
 /*--------------------------------------------------------------------
- * NB: This function is called from the test-load subprocess.
+ * NB: This function is called in/from the test-load subprocess.
  */
 
 int
@@ -440,51 +563,52 @@ VCL_TestLoad(const char *fn)
 
 /*--------------------------------------------------------------------*/
 
-static void
-vcl_print_refs(VRT_CTX)
+static struct vsb *
+vcl_print_refs(const struct vcl *vcl)
 {
-	struct vcl *vcl;
+	struct vsb *msg;
 	struct vclref *ref;
 
-	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
-	CHECK_OBJ_NOTNULL(ctx->vcl, VCL_MAGIC);
-	AN(ctx->msg);
-	vcl = ctx->vcl;
-	VSB_printf(ctx->msg, "VCL %s is waiting for:", vcl->loaded_name);
+	CHECK_OBJ_NOTNULL(vcl, VCL_MAGIC);
+	msg = VSB_new_auto();
+	AN(msg);
+
+	VSB_printf(msg, "VCL %s is waiting for:", vcl->loaded_name);
 	Lck_Lock(&vcl_mtx);
-	VTAILQ_FOREACH(ref, &ctx->vcl->ref_list, list)
-		VSB_printf(ctx->msg, "\n\t- %s", ref->desc);
+	VTAILQ_FOREACH(ref, &vcl->ref_list, list)
+		VSB_printf(msg, "\n\t- %s", ref->desc);
 	Lck_Unlock(&vcl_mtx);
+	AZ(VSB_finish(msg));
+	return (msg);
 }
 
 static int
-vcl_set_state(VRT_CTX, const char *state)
+vcl_set_state(struct vcl *vcl, const char *state, struct vsb **msg)
 {
-	struct vcl *vcl;
+	struct vsb *nomsg = NULL;
 	int i = 0;
 
 	ASSERT_CLI();
-	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
-	CHECK_OBJ_NOTNULL(ctx->vcl, VCL_MAGIC);
-	CHECK_OBJ_NOTNULL(ctx->vcl->conf, VCL_CONF_MAGIC);
-	AN(ctx->handling);
-	AN(ctx->vcl);
-	AN(state);
-	assert(ctx->msg != NULL || *state == '0');
 
-	vcl = ctx->vcl;
-	AZ(errno=pthread_rwlock_wrlock(&vcl->temp_rwl));
+	CHECK_OBJ_NOTNULL(vcl, VCL_MAGIC);
+	AN(state);
+	AN(msg);
+	AZ(*msg);
+
 	AN(vcl->temp);
 
 	switch (state[0]) {
 	case '0':
-		assert(vcl->temp != VCL_TEMP_COLD);
-		if (vcl->busy == 0 && VCL_WARM(vcl)) {
-
+		if (vcl->temp == VCL_TEMP_COLD)
+			break;
+		if (vcl->busy == 0 && vcl->temp->is_warm) {
+			Lck_Lock(&vcl_mtx);
 			vcl->temp = VTAILQ_EMPTY(&vcl->ref_list) ?
 			    VCL_TEMP_COLD : VCL_TEMP_COOLING;
-			AZ(vcl_send_event(ctx, VCL_EVENT_COLD));
+			Lck_Unlock(&vcl_mtx);
 			vcl_BackendEvent(vcl, VCL_EVENT_COLD);
+			AZ(vcl_send_event(vcl, VCL_EVENT_COLD, msg));
+			AZ(*msg);
 		}
 		else if (vcl->busy)
 			vcl->temp = VCL_TEMP_BUSY;
@@ -494,103 +618,111 @@ vcl_set_state(VRT_CTX, const char *state)
 			vcl->temp = VCL_TEMP_COOLING;
 		break;
 	case '1':
-		assert(vcl->temp != VCL_TEMP_WARM);
+		if (vcl->temp == VCL_TEMP_WARM)
+			break;
 		/* The warm VCL hasn't seen a cold event yet */
 		if (vcl->temp == VCL_TEMP_BUSY)
 			vcl->temp = VCL_TEMP_WARM;
 		/* The VCL must first reach a stable cold state */
 		else if (vcl->temp == VCL_TEMP_COOLING) {
-			vcl_print_refs(ctx);
+			*msg = vcl_print_refs(vcl);
 			i = -1;
 		}
 		else {
+			Lck_Lock(&vcl_mtx);
 			vcl->temp = VCL_TEMP_WARM;
-			i = vcl_send_event(ctx, VCL_EVENT_WARM);
-			if (i == 0)
+			Lck_Unlock(&vcl_mtx);
+			i = vcl_send_event(vcl, VCL_EVENT_WARM, msg);
+			if (i == 0) {
 				vcl_BackendEvent(vcl, VCL_EVENT_WARM);
-			else
-				AZ(vcl->conf->event_vcl(ctx, VCL_EVENT_COLD));
+				break;
+			}
+			AZ(vcl_send_event(vcl, VCL_EVENT_COLD, &nomsg));
+			AZ(nomsg);
+			vcl->temp = VCL_TEMP_COLD;
 		}
 		break;
 	default:
 		WRONG("Wrong enum state");
 	}
-	AZ(errno=pthread_rwlock_unlock(&vcl->temp_rwl));
+	if (i == 0 && state[1])
+		bprintf(vcl->state, "%s", state + 1);
 
 	return (i);
 }
 
 static void
-vcl_cancel_load(VRT_CTX, struct cli *cli, const char *name, const char *step)
+vcl_cancel_load(struct vcl *vcl, struct cli *cli, struct vsb *msg,
+    const char *name, const char *step)
 {
-	struct vcl *vcl = ctx->vcl;
 
 	CHECK_OBJ_NOTNULL(vcl, VCL_MAGIC);
 	CHECK_OBJ_NOTNULL(vcl->conf, VCL_CONF_MAGIC);
 
-	AZ(VSB_finish(ctx->msg));
 	VCLI_SetResult(cli, CLIS_CANT);
 	VCLI_Out(cli, "VCL \"%s\" Failed %s", name, step);
-	if (VSB_len(ctx->msg))
-		VCLI_Out(cli, "\nMessage:\n\t%s", VSB_data(ctx->msg));
-	*ctx->handling = 0;
-	AZ(vcl_send_event(ctx, VCL_EVENT_DISCARD));
+	if (VSB_len(msg))
+		VCLI_Out(cli, "\nMessage:\n\t%s", VSB_data(msg));
+	VSB_destroy(&msg);
+
+	AZ(vcl_send_event(vcl, VCL_EVENT_DISCARD, &msg));
+	AZ(msg);
+
 	vcl_KillBackends(vcl);
 	free(vcl->loaded_name);
 	VCL_Close(&vcl);
 }
 
 static void
-vcl_load(struct cli *cli, struct vrt_ctx *ctx,
+vcl_load(struct cli *cli,
     const char *name, const char *fn, const char *state)
 {
 	struct vcl *vcl;
-	int i;
+	struct vsb *msg;
 
 	ASSERT_CLI();
+	ASSERT_VCL_ACTIVE();
 
 	vcl = vcl_find(name);
 	AZ(vcl);
 
-	vcl = VCL_Open(fn, ctx->msg);
+	msg = VSB_new_auto();
+	AN(msg);
+	vcl = VCL_Open(fn, msg);
+	AZ(VSB_finish(msg));
 	if (vcl == NULL) {
-		AZ(VSB_finish(ctx->msg));
 		VCLI_SetResult(cli, CLIS_PARAM);
-		VCLI_Out(cli, "%s", VSB_data(ctx->msg));
+		VCLI_Out(cli, "%s", VSB_data(msg));
+		VSB_destroy(&msg);
 		return;
 	}
 
+	VSB_destroy(&msg);
+
 	vcl->loaded_name = strdup(name);
 	XXXAN(vcl->loaded_name);
 	VTAILQ_INIT(&vcl->director_list);
 	VTAILQ_INIT(&vcl->ref_list);
-	VTAILQ_INIT(&vcl->vfps);
+	VTAILQ_INIT(&vcl->filters);
 
 	vcl->temp = VCL_TEMP_INIT;
 
-	ctx->vcl = vcl;
+	if (vcl_send_event(vcl, VCL_EVENT_LOAD, &msg)) {
+		vcl_cancel_load(vcl, cli, msg, name, "initialization");
+		return;
+	}
+	VSB_destroy(&msg);
 
-	VSB_clear(ctx->msg);
-	i = vcl_send_event(ctx, VCL_EVENT_LOAD);
-	if (i || *ctx->handling != VCL_RET_OK) {
-		vcl_cancel_load(ctx, cli, name, "initialization");
-		return;
-	}
-	assert(*ctx->handling == VCL_RET_OK);
-	VSB_clear(ctx->msg);
-	i = vcl_set_state(ctx, state);
-	if (i) {
+	if (vcl_set_state(vcl, state, &msg)) {
 		assert(*state == '1');
-		vcl_cancel_load(ctx, cli, name, "warmup");
+		vcl_cancel_load(vcl, cli, msg, name, "warmup");
 		return;
 	}
-	bprintf(vcl->state, "%s", state + 1);
+	if (msg)
+		VSB_destroy(&msg);
+
 	VCLI_Out(cli, "Loaded \"%s\" as \"%s\"", fn , name);
 	VTAILQ_INSERT_TAIL(&vcl_head, vcl, list);
-	Lck_Lock(&vcl_mtx);
-	if (vcl_active == NULL)
-		vcl_active = vcl;
-	Lck_Unlock(&vcl_mtx);
 	VSC_C_main->n_vcl++;
 	VSC_C_main->n_vcl_avail++;
 }
@@ -600,28 +732,23 @@ vcl_load(struct cli *cli, struct vrt_ctx *ctx,
 void
 VCL_Poll(void)
 {
-	struct vrt_ctx *ctx;
+	struct vsb *nomsg = NULL;
 	struct vcl *vcl, *vcl2;
 
 	ASSERT_CLI();
-	ctx = vcl_get_ctx(0, 0);
+	ASSERT_VCL_ACTIVE();
 	VTAILQ_FOREACH_SAFE(vcl, &vcl_head, list, vcl2) {
 		if (vcl->temp == VCL_TEMP_BUSY ||
-		    vcl->temp == VCL_TEMP_COOLING) {
-			ctx->vcl = vcl;
-			ctx->syntax = ctx->vcl->conf->syntax;
-			ctx->method = 0;
-			(void)vcl_set_state(ctx, "0");
-		}
+		    vcl->temp == VCL_TEMP_COOLING)
+			AZ(vcl_set_state(vcl, "0", &nomsg));
+		AZ(nomsg);
 		if (vcl->discard && vcl->temp == VCL_TEMP_COLD) {
 			AZ(vcl->busy);
 			assert(vcl != vcl_active);
 			assert(VTAILQ_EMPTY(&vcl->ref_list));
 			VTAILQ_REMOVE(&vcl_head, vcl, list);
-			ctx->method = VCL_MET_FINI;
-			ctx->vcl = vcl;
-			ctx->syntax = ctx->vcl->conf->syntax;
-			AZ(vcl_send_event(ctx, VCL_EVENT_DISCARD));
+			AZ(vcl_send_event(vcl, VCL_EVENT_DISCARD, &nomsg));
+			AZ(nomsg);
 			vcl_KillBackends(vcl);
 			free(vcl->loaded_name);
 			VCL_Close(&vcl);
@@ -629,7 +756,6 @@ VCL_Poll(void)
 			VSC_C_main->n_vcl_discard--;
 		}
 	}
-	vcl_rel_ctx(&ctx);
 }
 
 /*--------------------------------------------------------------------*/
@@ -639,12 +765,16 @@ vcl_cli_list(struct cli *cli, const char * const *av, void *priv)
 {
 	struct vcl *vcl;
 	const char *flg;
+	struct vte *vte;
 
 	/* NB: Shall generate same output as mcf_vcl_list() */
 
 	(void)av;
 	(void)priv;
 	ASSERT_CLI();
+	ASSERT_VCL_ACTIVE();
+	vte = VTE_new(7, 80);
+	AN(vte);
 	VTAILQ_FOREACH(vcl, &vcl_head, list) {
 		if (vcl == vcl_active) {
 			flg = "active";
@@ -652,56 +782,107 @@ vcl_cli_list(struct cli *cli, const char * const *av, void *priv)
 			flg = "discarded";
 		} else
 			flg = "available";
-		VCLI_Out(cli, "%-10s %5s/%-8s %6u %s",
-		    flg, vcl->state, vcl->temp, vcl->busy, vcl->loaded_name);
+		VTE_printf(vte, "%s\t%s\t%s\t\v%u\t%s", flg, vcl->state,
+		    vcl->temp->name, vcl->busy, vcl->loaded_name);
 		if (vcl->label != NULL) {
-			VCLI_Out(cli, " -> %s", vcl->label->loaded_name);
+			VTE_printf(vte, "\t->\t%s", vcl->label->loaded_name);
 			if (vcl->nrefs)
-				VCLI_Out(cli, " (%d return(vcl)%s)",
+				VTE_printf(vte, " (%d return(vcl)%s)",
 				    vcl->nrefs, vcl->nrefs > 1 ? "'s" : "");
 		} else if (vcl->nlabels > 0) {
-			VCLI_Out(cli, " (%d label%s)",
+			VTE_printf(vte, "\t<-\t(%d label%s)",
 			    vcl->nlabels, vcl->nlabels > 1 ? "s" : "");
 		}
-		VCLI_Out(cli, "\n");
+		VTE_cat(vte, "\n");
 	}
+	AZ(VTE_finish(vte));
+	AZ(VTE_format(vte, VCLI_VTE_format, cli));
+	VTE_destroy(&vte);
+}
+
+static void v_matchproto_(cli_func_t)
+vcl_cli_list_json(struct cli *cli, const char * const *av, void *priv)
+{
+	struct vcl *vcl;
+
+	(void)priv;
+	ASSERT_CLI();
+	ASSERT_VCL_ACTIVE();
+	VCLI_JSON_begin(cli, 2, av);
+	VCLI_Out(cli, ",\n");
+	VTAILQ_FOREACH(vcl, &vcl_head, list) {
+		VCLI_Out(cli, "{\n");
+		VSB_indent(cli->sb, 2);
+		VCLI_Out(cli, "\"status\": ");
+		if (vcl == vcl_active) {
+			VCLI_Out(cli, "\"active\",\n");
+		} else if (vcl->discard) {
+			VCLI_Out(cli, "\"discarded\",\n");
+		} else
+			VCLI_Out(cli, "\"available\",\n");
+		VCLI_Out(cli, "\"state\": \"%s\",\n", vcl->state);
+		VCLI_Out(cli, "\"temperature\": \"%s\",\n", vcl->temp->name);
+		VCLI_Out(cli, "\"busy\": %u,\n", vcl->busy);
+		VCLI_Out(cli, "\"name\": \"%s\"", vcl->loaded_name);
+		if (vcl->label != NULL) {
+			VCLI_Out(cli, ",\n");
+			VCLI_Out(cli, "\"label\": {\n");
+			VSB_indent(cli->sb, 2);
+				VCLI_Out(cli, "\"name\": \"%s\"",
+					 vcl->label->loaded_name);
+			if (vcl->nrefs)
+				VCLI_Out(cli, ",\n\"refs\": %d", vcl->nrefs);
+			VCLI_Out(cli, "\n");
+			VCLI_Out(cli, "}");
+			VSB_indent(cli->sb, -2);
+		} else if (vcl->nlabels > 0) {
+			VCLI_Out(cli, ",\n");
+			VCLI_Out(cli, "\"labels\": %d", vcl->nlabels);
+		}
+		VSB_indent(cli->sb, -2);
+		VCLI_Out(cli, "\n}");
+		if (VTAILQ_NEXT(vcl, list) != NULL)
+			VCLI_Out(cli, ",\n");
+	}
+	VCLI_JSON_end(cli);
 }
 
 static void v_matchproto_(cli_func_t)
 vcl_cli_load(struct cli *cli, const char * const *av, void *priv)
 {
-	struct vrt_ctx *ctx;
 
 	AZ(priv);
 	ASSERT_CLI();
-	ctx = vcl_get_ctx(VCL_MET_INIT, 1);
-	vcl_load(cli, ctx, av[2], av[3], av[4]);
-	vcl_rel_ctx(&ctx);
+	// XXX move back code from vcl_load?
+	vcl_load(cli, av[2], av[3], av[4]);
 }
 
 static void v_matchproto_(cli_func_t)
 vcl_cli_state(struct cli *cli, const char * const *av, void *priv)
 {
-	struct vrt_ctx *ctx;
+	struct vcl *vcl;
+	struct vsb *msg = NULL;
 
 	AZ(priv);
 	ASSERT_CLI();
+	ASSERT_VCL_ACTIVE();
 	AN(av[2]);
 	AN(av[3]);
-	ctx = vcl_get_ctx(0, 1);
-	ctx->vcl = vcl_find(av[2]);
-	AN(ctx->vcl);			// MGT ensures this
-	if (vcl_set_state(ctx, av[3]) == 0) {
-		bprintf(ctx->vcl->state, "%s", av[3] + 1);
-	} else {
-		AZ(VSB_finish(ctx->msg));
+
+	vcl = vcl_find(av[2]);
+	AN(vcl);
+
+	if (vcl_set_state(vcl, av[3], &msg)) {
+		CHECK_OBJ_NOTNULL(msg, VSB_MAGIC);
+
 		VCLI_SetResult(cli, CLIS_CANT);
-		VCLI_Out(cli, "Failed <vcl.state %s %s>", ctx->vcl->loaded_name,
+		VCLI_Out(cli, "Failed <vcl.state %s %s>", vcl->loaded_name,
 		    av[3] + 1);
-		if (VSB_len(ctx->msg))
-			VCLI_Out(cli, "\nMessage:\n\t%s", VSB_data(ctx->msg));
+		if (VSB_len(msg))
+			VCLI_Out(cli, "\nMessage:\n\t%s", VSB_data(msg));
 	}
-	vcl_rel_ctx(&ctx);
+	if (msg)
+		VSB_destroy(&msg);
 }
 
 static void v_matchproto_(cli_func_t)
@@ -710,10 +891,11 @@ vcl_cli_discard(struct cli *cli, const char * const *av, void *priv)
 	struct vcl *vcl;
 
 	ASSERT_CLI();
+	ASSERT_VCL_ACTIVE();
 	(void)cli;
 	AZ(priv);
 	vcl = vcl_find(av[2]);
-	AN(vcl);			// MGT ensures this
+	CHECK_OBJ_NOTNULL(vcl, VCL_MAGIC);		// MGT ensures this
 	Lck_Lock(&vcl_mtx);
 	assert (vcl != vcl_active);	// MGT ensures this
 	AZ(vcl->nlabels);		// MGT ensures this
@@ -721,16 +903,15 @@ vcl_cli_discard(struct cli *cli, const char * const *av, void *priv)
 	VSC_C_main->n_vcl_avail--;
 	vcl->discard = 1;
 	if (vcl->label != NULL) {
-		AZ(strcmp(vcl->state, VCL_TEMP_LABEL));
+		AZ(strcmp(vcl->state, VCL_TEMP_LABEL->name));
 		vcl->label->nlabels--;
 		vcl->label= NULL;
 	}
 	Lck_Unlock(&vcl_mtx);
 
-	if (!strcmp(vcl->state, VCL_TEMP_LABEL)) {
+	if (!strcmp(vcl->state, VCL_TEMP_LABEL->name)) {
 		VTAILQ_REMOVE(&vcl_head, vcl, list);
 		free(vcl->loaded_name);
-		AZ(errno=pthread_rwlock_destroy(&vcl->temp_rwl));
 		FREE_OBJ(vcl);
 	} else if (vcl->temp == VCL_TEMP_COLD) {
 		VCL_Poll();
@@ -744,6 +925,7 @@ vcl_cli_label(struct cli *cli, const char * const *av, void *priv)
 	struct vcl *vcl;
 
 	ASSERT_CLI();
+	ASSERT_VCL_ACTIVE();
 	(void)cli;
 	(void)priv;
 	vcl = vcl_find(av[3]);
@@ -752,10 +934,9 @@ vcl_cli_label(struct cli *cli, const char * const *av, void *priv)
 	if (lbl == NULL) {
 		ALLOC_OBJ(lbl, VCL_MAGIC);
 		AN(lbl);
-		bprintf(lbl->state, "%s", VCL_TEMP_LABEL);
+		bprintf(lbl->state, "%s", VCL_TEMP_LABEL->name);
 		lbl->temp = VCL_TEMP_WARM;
 		REPLACE(lbl->loaded_name, av[2]);
-		AZ(errno=pthread_rwlock_init(&lbl->temp_rwl, NULL));
 		VTAILQ_INSERT_TAIL(&vcl_head, lbl, list);
 	}
 	if (lbl->label != NULL)
@@ -770,6 +951,7 @@ vcl_cli_use(struct cli *cli, const char * const *av, void *priv)
 	struct vcl *vcl;
 
 	ASSERT_CLI();
+	ASSERT_VCL_ACTIVE();
 	AN(cli);
 	AZ(priv);
 	vcl = vcl_find(av[2]);
@@ -785,38 +967,50 @@ vcl_cli_show(struct cli *cli, const char * const *av, void *priv)
 {
 	struct vcl *vcl;
 	int verbose = 0;
-	int i;
+	int i = 2;
+	unsigned u;
 
 	ASSERT_CLI();
+	ASSERT_VCL_ACTIVE();
 	AZ(priv);
-	if (!strcmp(av[2], "-v") && av[3] == NULL) {
-		VCLI_Out(cli, "Too few parameters");
-		VCLI_SetResult(cli, CLIS_TOOFEW);
-		return;
-	} else if (strcmp(av[2], "-v") && av[3] != NULL) {
-		VCLI_Out(cli, "Unknown options '%s'", av[2]);
+
+	if (av[i] != NULL && !strcmp(av[i], "-v")) {
+		verbose = 1;
+		i++;
+	}
+
+	if (av[i] == NULL) {
+		vcl = vcl_active;
+		AN(vcl);
+	} else {
+		vcl = vcl_find(av[i]);
+		i++;
+	}
+
+	if (av[i] != NULL) {
+		VCLI_Out(cli, "Too many parameters: '%s'", av[i]);
 		VCLI_SetResult(cli, CLIS_PARAM);
 		return;
-	} else if (av[3] != NULL) {
-		verbose = 1;
-		vcl = vcl_find(av[3]);
-	} else
-		vcl = vcl_find(av[2]);
+	}
 
 	if (vcl == NULL) {
-		VCLI_Out(cli, "No VCL named '%s'",
-		    av[3] == NULL ? av[2] : av[3]);
+		VCLI_Out(cli, "No VCL named '%s'", av[i - 1]);
 		VCLI_SetResult(cli, CLIS_PARAM);
 		return;
 	}
 	CHECK_OBJ_NOTNULL(vcl, VCL_MAGIC);
+	if (vcl->label) {
+		vcl = vcl->label;
+		CHECK_OBJ_NOTNULL(vcl, VCL_MAGIC);
+		AZ(vcl->label);
+	}
 	CHECK_OBJ_NOTNULL(vcl->conf, VCL_CONF_MAGIC);
 	if (verbose) {
-		for (i = 0; i < vcl->conf->nsrc; i++)
-			VCLI_Out(cli, "// VCL.SHOW %d %zd %s\n%s\n",
-			    i, strlen(vcl->conf->srcbody[i]),
-			    vcl->conf->srcname[i],
-			    vcl->conf->srcbody[i]);
+		for (u = 0; u < vcl->conf->nsrc; u++)
+			VCLI_Out(cli, "// VCL.SHOW %u %zd %s\n%s\n",
+			    u, strlen(vcl->conf->srcbody[u]),
+			    vcl->conf->srcname[u],
+			    vcl->conf->srcbody[u]);
 	} else {
 		VCLI_Out(cli, "%s", vcl->conf->srcbody[0]);
 	}
@@ -826,7 +1020,7 @@ vcl_cli_show(struct cli *cli, const char * const *av, void *priv)
 
 static struct cli_proto vcl_cmds[] = {
 	{ CLICMD_VCL_LOAD,		"", vcl_cli_load },
-	{ CLICMD_VCL_LIST,		"", vcl_cli_list },
+	{ CLICMD_VCL_LIST,		"", vcl_cli_list, vcl_cli_list_json },
 	{ CLICMD_VCL_STATE,		"", vcl_cli_state },
 	{ CLICMD_VCL_DISCARD,		"", vcl_cli_discard },
 	{ CLICMD_VCL_USE,		"", vcl_cli_use },
@@ -845,4 +1039,5 @@ VCL_Init(void)
 	ws_snapshot_cli = WS_Snapshot(&ws_cli);
 	CLI_AddFuncs(vcl_cmds);
 	Lck_New(&vcl_mtx, lck_vcl);
+	VSL_Setup(&vsl_cli, NULL, 0);
 }
