@@ -4,6 +4,8 @@
  *
  * Author: Dridi Boukelmoune <dridi@varnish-software.com>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -34,7 +36,6 @@
 #include <sys/ioctl.h>		// Linux: struct winsize
 
 #include <ctype.h>
-#include <errno.h>
 #include <fcntl.h>
 #include <inttypes.h>
 #include <poll.h>
@@ -49,6 +50,7 @@
 
 #include "vtc.h"
 
+#include "vre.h"
 #include "vev.h"
 #include "vlu.h"
 #include "vsb.h"
@@ -227,7 +229,7 @@ term_resize(struct process *pp, int lin, int col)
 	vram = calloc(lin, sizeof *pp->vram);
 	AN(vram);
 	for (i = 0; i < lin; i++) {
-		vram[i] = malloc(col + 1L);
+		vram[i] = calloc(col + 1L, 1);
 		AN(vram[i]);
 		memset(vram[i], ' ', col);
 		vram[i][col] = '\0';
@@ -255,9 +257,10 @@ term_resize(struct process *pp, int lin, int col)
 }
 
 static int
-term_match_textline(const struct process *pp, int *x, int y, const char *pat)
+term_find_textline(const struct process *pp, int *x, int y, const char *pat)
 {
 	const char *t;
+	int l;
 
 	if (*x == 0) {
 		t = strstr(pp->vram[y], pat);
@@ -267,26 +270,28 @@ term_match_textline(const struct process *pp, int *x, int y, const char *pat)
 		}
 	} else if (*x <= pp->ncol) {
 		t = pp->vram[y] + *x - 1;
-		if (!memcmp(t, pat, strlen(pat)))
+		l = strlen(pat);
+		assert((*x - 1) + (l - 1) < pp->ncol);
+		if (!memcmp(t, pat, l))
 			return (1);
 	}
 	return (0);
 }
 
 static int
-term_match_text(const struct process *pp, int *x, int *y, const char *pat)
+term_find_text(const struct process *pp, int *x, int *y, const char *pat)
 {
 	int yy;
 
 	if (*y == 0) {
 		for (yy = 0; yy < pp->nlin; yy++) {
-			if (term_match_textline(pp, x, yy, pat)) {
+			if (term_find_textline(pp, x, yy, pat)) {
 				*y = yy + 1;
 				return (1);
 			}
 		}
 	} else if (*y <= pp->nlin) {
-		if (term_match_textline(pp, x, *y - 1, pat))
+		if (term_find_textline(pp, x, *y - 1, pat))
 			return (1);
 	}
 	return (0);
@@ -300,34 +305,48 @@ term_expect_text(struct process *pp,
 	char *t;
 
 	y = strtoul(lin, NULL, 0);
+	if (y < 0 || y > pp->nlin)
+		vtc_fatal(pp->vl, "YYY %d nlin %d", y, pp->nlin);
 	x = strtoul(col, NULL, 0);
+	for(l = 0; l <= 10 && x > pp->ncol; l++)	// wait for screen change
+		usleep(100000);
+	if (x < 0 || x > pp->ncol)
+		vtc_fatal(pp->vl, "XXX %d ncol %d", x, pp->ncol);
 	l = strlen(pat);
-	AZ(pthread_mutex_lock(&pp->mtx));
-	while (!term_match_text(pp, &x, &y, pat)) {
+	if (x + l - 1 > pp->ncol)
+		vtc_fatal(pp->vl, "XXX %d ncol %d", x + l - 1, pp->ncol);
+	PTOK(pthread_mutex_lock(&pp->mtx));
+	while (!term_find_text(pp, &x, &y, pat)) {
 		if (x != 0 && y != 0) {
 			t = pp->vram[y - 1] + x - 1;
 			vtc_log(pp->vl, 4,
 			    "text at %d,%d: '%.*s'", y, x, l, t);
 		}
-		AZ(pthread_mutex_unlock(&pp->mtx));
+		PTOK(pthread_mutex_unlock(&pp->mtx));
 		usleep(d);
-		AZ(pthread_mutex_lock(&pp->mtx));
-		if (d < 300000)
+		PTOK(pthread_mutex_lock(&pp->mtx));
+		if (d < 3000000)
 			d += d;
 	}
-	AZ(pthread_mutex_unlock(&pp->mtx));
+	PTOK(pthread_mutex_unlock(&pp->mtx));
 	vtc_log(pp->vl, 4, "found expected text at %d,%d: '%s'", y, x, pat);
 }
 
 static void
 term_expect_cursor(const struct process *pp, const char *lin, const char *col)
 {
-	int x, y;
+	int x, y, l;
 	const teken_pos_t *pos;
 
 	pos = teken_get_cursor(pp->tek);
 	y = strtoul(lin, NULL, 0);
+	if (y < 0 || y > pp->nlin)
+		vtc_fatal(pp->vl, "YYY %d nlin %d", y, pp->nlin);
 	x = strtoul(col, NULL, 0);
+	for(l = 0; l < 10 && x > pp->ncol; l++)	// wait for screen change
+		usleep(100000);
+	if (x < 0 || x > pp->ncol)
+		vtc_fatal(pp->vl, "XXX %d ncol %d", x, pp->ncol);
 	if (y != 0 && (y-1) != pos->tp_row)
 		vtc_fatal(pp->vl, "Cursor on line %d (expected %d)",
 		    pos->tp_row + 1, y);
@@ -336,6 +355,65 @@ term_expect_cursor(const struct process *pp, const char *lin, const char *col)
 		    pos->tp_col + 1, y);
 }
 
+static void
+term_match_text(struct process *pp,
+    const char *lin, const char *col, const char *re)
+{
+	int i, l, err, erroff;
+	struct vsb *vsb, re_vsb[1];
+	size_t len;
+	ssize_t x, y;
+	vre_t *vre;
+	char errbuf[VRE_ERROR_LEN];
+
+	vsb = VSB_new_auto();
+	AN(vsb);
+
+	y = strtoul(lin, NULL, 0);
+	if (y < 0 || y > pp->nlin)
+		vtc_fatal(pp->vl, "YYY %zd nlin %d", y, pp->nlin);
+	x = strtoul(col, NULL, 0);
+	for(l = 0; l < 10 && x > pp->ncol; l++)	// wait for screen change
+		usleep(100000);
+	if (x < 0 || x > pp->ncol)
+		vtc_fatal(pp->vl, "XXX %zd ncol %d", x, pp->ncol);
+
+	if (x)
+		x--;
+
+	if (y)
+		y--;
+
+	vre = VRE_compile(re, 0, &err, &erroff, 1);
+	if (vre == NULL) {
+		AN(VSB_init(re_vsb, errbuf, sizeof errbuf));
+		AZ(VRE_error(re_vsb, err));
+		AZ(VSB_finish(re_vsb));
+		VSB_fini(re_vsb);
+		vtc_fatal(pp->vl, "invalid regexp \"%s\" at %d (%s)",
+		    re, erroff, errbuf);
+	}
+
+	PTOK(pthread_mutex_lock(&pp->mtx));
+
+	len = (pp->nlin - y) * (pp->ncol - x);
+	for (i = y; i < pp->nlin; i++) {
+		VSB_bcat(vsb, &pp->vram[i][x], pp->ncol - x);
+		VSB_putc(vsb, '\n');
+	}
+
+	AZ(VSB_finish(vsb));
+
+	if (VRE_match(vre, VSB_data(vsb), len, 0, NULL) < 1)
+		vtc_fatal(pp->vl, "match failed: (\"%s\")", re);
+	else
+		vtc_log(pp->vl, 4, "match succeeded");
+
+	PTOK(pthread_mutex_unlock(&pp->mtx));
+	VSB_destroy(&vsb);
+	VRE_free(&vre);
+}
+
 /**********************************************************************
  * Allocate and initialize a process
  */
@@ -383,9 +461,9 @@ process_new(const char *name)
 	ALLOC_OBJ(p, PROCESS_MAGIC);
 	AN(p);
 	REPLACE(p->name, name);
-	AZ(pthread_mutex_init(&p->mtx, NULL));
+	PTOK(pthread_mutex_init(&p->mtx, NULL));
 
-	p->vl = vtc_logopen(name);
+	p->vl = vtc_logopen("%s", name);
 	AN(p->vl);
 
 	PROCESS_EXPAND(dir, "${tmpdir}/%s", name);
@@ -417,7 +495,7 @@ process_delete(struct process *p)
 	int i;
 
 	CHECK_OBJ_NOTNULL(p, PROCESS_MAGIC);
-	AZ(pthread_mutex_destroy(&p->mtx));
+	PTOK(pthread_mutex_destroy(&p->mtx));
 	vtc_logclose(p->vl);
 	free(p->name);
 	free(p->dir);
@@ -476,9 +554,9 @@ process_stdout(const struct vev *ev, int what)
 		vtc_log(p->vl, 4, "stdout read %d", i);
 		return (1);
 	}
-	AZ(pthread_mutex_lock(&p->mtx));
+	PTOK(pthread_mutex_lock(&p->mtx));
 	p->stdout_bytes += i;
-	AZ(pthread_mutex_unlock(&p->mtx));
+	PTOK(pthread_mutex_unlock(&p->mtx));
 	if (p->log == 1)
 		(void)VLU_Feed(p->vlu_stdout, buf, i);
 	else if (p->log == 2)
@@ -486,9 +564,9 @@ process_stdout(const struct vev *ev, int what)
 	else if (p->log == 3)
 		vtc_hexdump(p->vl, 4, "stdout", buf, i);
 	assert(write(p->f_stdout, buf, i) == i);
-	AZ(pthread_mutex_lock(&p->mtx));
+	PTOK(pthread_mutex_lock(&p->mtx));
 	teken_input(p->tek, buf, i);
-	AZ(pthread_mutex_unlock(&p->mtx));
+	PTOK(pthread_mutex_unlock(&p->mtx));
 	return (0);
 }
 
@@ -506,14 +584,21 @@ process_stderr(const struct vev *ev, int what)
 		vtc_log(p->vl, 4, "stderr read %d", i);
 		return (1);
 	}
-	AZ(pthread_mutex_lock(&p->mtx));
+	PTOK(pthread_mutex_lock(&p->mtx));
 	p->stderr_bytes += i;
-	AZ(pthread_mutex_unlock(&p->mtx));
+	PTOK(pthread_mutex_unlock(&p->mtx));
 	vtc_dump(p->vl, 4, "stderr", buf, i);
 	assert(write(p->f_stderr, buf, i) == i);
 	return (0);
 }
 
+static void
+process_cleanup(void *priv)
+{
+	struct vev_root *evb = priv;
+	VEV_Destroy(&evb);
+}
+
 static void *
 process_thread(void *priv)
 {
@@ -531,6 +616,7 @@ process_thread(void *priv)
 
 	evb = VEV_New();
 	AN(evb);
+	pthread_cleanup_push(process_cleanup, evb);
 
 	ev = VEV_Alloc();
 	AN(ev);
@@ -568,14 +654,15 @@ process_thread(void *priv)
 	closefd(&p->f_stdout);
 	closefd(&p->f_stderr);
 
-	AZ(pthread_mutex_lock(&p->mtx));
+	PTOK(pthread_mutex_lock(&p->mtx));
 
 	/* NB: We keep the other macros around */
 	macro_undef(p->vl, p->name, "pid");
 	p->pid = -1;
 
-	AZ(pthread_mutex_unlock(&p->mtx));
+	PTOK(pthread_mutex_unlock(&p->mtx));
 
+	pthread_cleanup_pop(0);
 	VEV_Destroy(&evb);
 	if (p->log == 1) {
 		VLU_Destroy(&p->vlu_stdout);
@@ -700,7 +787,7 @@ process_start(struct process *p)
 	macro_def(p->vl, p->name, "out", "%s", p->out);
 	macro_def(p->vl, p->name, "err", "%s", p->err);
 	p->hasthread = 1;
-	AZ(pthread_create(&p->tp, NULL, process_thread, p));
+	PTOK(pthread_create(&p->tp, NULL, process_thread, p));
 }
 
 /**********************************************************************
@@ -713,7 +800,7 @@ process_wait(struct process *p)
 	void *v;
 
 	if (p->hasthread) {
-		AZ(pthread_join(p->tp, &v));
+		PTOK(pthread_join(p->tp, &v));
 		p->hasthread = 0;
 	}
 	vtc_log(p->vl, 4, "stdout %ju bytes, stderr %ju bytes",
@@ -733,9 +820,9 @@ process_kill(struct process *p, const char *sig)
 	CHECK_OBJ_NOTNULL(p, PROCESS_MAGIC);
 	AN(sig);
 
-	AZ(pthread_mutex_lock(&p->mtx));
+	PTOK(pthread_mutex_lock(&p->mtx));
 	pid = p->pid;
-	AZ(pthread_mutex_unlock(&p->mtx));
+	PTOK(pthread_mutex_unlock(&p->mtx));
 
 	if (pid <= 0)
 		vtc_fatal(p->vl, "Cannot signal a non-running process");
@@ -786,7 +873,6 @@ static void
 process_write_hex(const struct process *p, const char *text)
 {
 	struct vsb *vsb;
-	int j;
 
 	if (!p->hasthread)
 		vtc_fatal(p->vl, "Cannot write to a non-running process");
@@ -794,8 +880,7 @@ process_write_hex(const struct process *p, const char *text)
 	vsb = vtc_hex_to_bin(p->vl, text);
 	assert(VSB_len(vsb) >= 0);
 	vtc_hexdump(p->vl, 4, "sendhex", VSB_data(vsb), VSB_len(vsb));
-	j = write(p->fd_term, VSB_data(vsb), VSB_len(vsb));
-	assert(j == VSB_len(vsb));
+	AZ(VSB_tofile(vsb, p->fd_term));
 	VSB_destroy(&vsb);
 }
 
@@ -823,10 +908,14 @@ process_close(struct process *p)
  * Output from the stderr-pipe is copied verbatim to ${pNAME_err}, and
  * is always included in the vtc_log.
  *
- *	process pNAME SPEC [-log] [-dump] [-hexdump] [-expect-exit N]
- *		[-start] [-run]
- *		[-write STRING] [-writeln STRING]
- *		[-kill STRING] [-stop] [-wait] [-close]
+ *	process pNAME SPEC [-allow-core] [-expect-exit N] [-expect-signal N]
+ *		[-dump] [-hexdump] [-log]
+ *		[-run] [-close] [-kill SIGNAL] [-start] [-stop] [-wait]
+ *		[-write STRING] [-writeln STRING] [-writehex HEXSTRING]
+ *		[-need-bytes [+]NUMBER]
+ *		[-screen-dump] [-winsz LINES COLUMNS] [-ansi-response]
+ *		[-expect-cursor LINE COLUMN] [-expect-text LINE COLUMN TEXT]
+ *		[-match-text LINE COLUMN REGEXP]
  *
  * pNAME
  *	Name of the process. It must start with 'p'.
@@ -849,6 +938,12 @@ process_close(struct process *p)
  * \-expect-exit N
  *	Expect exit status N
  *
+ * \-expect-signal N
+ *	Expect signal in exit status N
+ *
+ * \-allow-core
+ *	Core dump in exit status is OK
+ *
  * \-wait
  *	Wait for the process to finish.
  *
@@ -856,20 +951,25 @@ process_close(struct process *p)
  *	Shorthand for -start -wait.
  *
  *	In most cases, if you just want to start a process and wait for it
- *	to finish, you can use the varnishtest ``shell`` command instead.
+ *	to finish, you can use the ``shell`` command instead.
  *	The following commands are equivalent::
  *
  *	    shell "do --something"
  *
  *	    process p1 "do --something" -run
  *
- *	However, you may use the the ``process`` variant to conveniently
+ *	However, you may use the ``process`` variant to conveniently
  *	collect the standard input and output without dealing with shell
  *	redirections yourself. The ``shell`` command can also expect an
  *	expression from either output, consider using it if you only need
  *	to match one.
  *
- * \-kill STRING
+ * \-key KEYSYM
+ *      Send emulated key-press.
+ *      KEYSYM can be one of (NPAGE, PPAGE, HOME, END)
+ *
+ *
+ * \-kill SIGNAL
  *	Send a signal to the process. The argument can be either
  *	the string "TERM", "INT", or "KILL" for SIGTERM, SIGINT or SIGKILL
  *	signals, respectively, or a hyphen (-) followed by the signal
@@ -885,6 +985,12 @@ process_close(struct process *p)
  * \-stop
  *	Shorthand for -kill TERM.
  *
+ * \-close
+ *	Alias for "-kill HUP"
+ *
+ * \-winsz LINES COLUMNS
+ *	Change the terminal window size to LIN lines and COL columns.
+ *
  * \-write STRING
  *	Write a string to the process' stdin.
  *
@@ -898,16 +1004,26 @@ process_close(struct process *p)
  *	Wait until at least NUMBER bytes have been received in total.
  *	If '+' is prefixed, NUMBER new bytes must be received.
  *
- * \-expect-text LIN COL PAT
- *	Wait for PAT to appear at LIN,COL on the virtual screen.
+ * \-ansi-response
+ *	Respond to terminal respond-back sequences
+ *
+ * \-expect-cursor LINE COLUMN
+ *	Expect cursors location
+ *
+ * \-expect-text LINE COLUMNS TEXT
+ *	Wait for TEXT to appear at LIN,COL on the virtual screen.
  *	Lines and columns are numbered 1...N
  *	LIN==0 means "on any line"
  *	COL==0 means "anywhere on the line"
  *
- * \-close
- *	Alias for "-kill HUP"
+ * \-match-text LINE COLUMN REGEXP
+ *	Wait for the PAT regular expression to match the text at LIN,COL on the virtual screen.
+ *	Lines and columns are numbered 1...N
+ *	LIN==0 means "on any line"
+ *	COL==0 means "anywhere on the line"
  *
- * \-screen_dump
+ *
+ * \-screen-dump
  *	Dump the virtual screen into vtc_log
  *
  */
@@ -921,7 +1037,6 @@ cmd_process(CMD_ARGS)
 	int spec_set = 0;
 
 	(void)priv;
-	(void)cmd;
 
 	if (av == NULL) {
 		/* Reset and free */
@@ -952,7 +1067,9 @@ cmd_process(CMD_ARGS)
 		p = process_new(av[0]);
 	av++;
 
+	PTOK(pthread_mutex_lock(&p->mtx));
 	bsnap = p->stdout_bytes;
+	PTOK(pthread_mutex_unlock(&p->mtx));
 
 	for (; *av != NULL; av++) {
 		if (vtc_error)
@@ -990,6 +1107,19 @@ cmd_process(CMD_ARGS)
 			p->log = 3;
 			continue;
 		}
+		if (!strcmp(*av, "-key")) {
+			if (!strcmp(av[1], "NPAGE"))
+				process_write(p, "\x1b\x5b\x36\x7e");
+			else if (!strcmp(av[1], "PPAGE"))
+				process_write(p, "\x1b\x5b\x35\x7e");
+			else if (!strcmp(av[1], "HOME"))
+				process_write(p, "\x1b\x4f\x48");
+			else if (!strcmp(av[1], "END"))
+				process_write(p, "\x1b\x4f\x46");
+			else
+				vtc_fatal(p->vl, "Unknown key %s", av[1]);
+			continue;
+		}
 		if (!strcmp(*av, "-kill")) {
 			process_kill(p, av[1]);
 			av++;
@@ -1008,9 +1138,9 @@ cmd_process(CMD_ARGS)
 				u += bsnap;
 			av++;
 			do {
-				AZ(pthread_mutex_lock(&p->mtx));
+				PTOK(pthread_mutex_lock(&p->mtx));
 				v = p->stdout_bytes;
-				AZ(pthread_mutex_unlock(&p->mtx));
+				PTOK(pthread_mutex_unlock(&p->mtx));
 				vtc_log(p->vl, 4, "Have %ju bytes", v);
 				usleep(500000);
 			} while(v < u);
@@ -1040,6 +1170,14 @@ cmd_process(CMD_ARGS)
 			av += 2;
 			continue;
 		}
+		if (!strcmp(*av, "-match-text")) {
+			AN(av[1]);
+			AN(av[2]);
+			AN(av[3]);
+			term_match_text(p, av[1], av[2], av[3]);
+			av += 3;
+			continue;
+		}
 		if (!strcmp(*av, "-screen_dump") ||
 		    !strcmp(*av, "-screen-dump")) {
 			term_screen_dump(p);
@@ -1064,10 +1202,11 @@ cmd_process(CMD_ARGS)
 			col = atoi(av[2]);
 			assert(col > 1);
 			av += 2;
-			AZ(pthread_mutex_lock(&p->mtx));
+			PTOK(pthread_mutex_lock(&p->mtx));
 			term_resize(p, lin, col);
-			AZ(pthread_mutex_unlock(&p->mtx));
+			PTOK(pthread_mutex_unlock(&p->mtx));
 			process_winsz(p, p->fd_term);
+			continue;
 		}
 		if (!strcmp(*av, "-write")) {
 			process_write(p, av[1]);
@@ -1086,8 +1225,7 @@ cmd_process(CMD_ARGS)
 			continue;
 		}
 		if (**av == '-' || spec_set)
-			vtc_fatal(p->vl, "Unknown process argument: %s",
-			    *av);
+			vtc_fatal(p->vl, "Unknown process argument: %s", *av);
 		REPLACE(p->spec, *av);
 		spec_set = 1;
 	}
